/*
 * Java
 *
 * Copyright 2014-2024 IS2T. All rights reserved
 * IS2T PROPRIETARY/CONFIDENTIAL Use is subject to license terms.
 */
package com.is2t.schedcontrol;

/**
 * A {@link ScheduleContext} is a scheduling policy that will be applied to a set of threads.
 * A thread can be associated to only one {@link ScheduleContext}.<p>
 * When a thread is running, the quantum counter of its {@link ScheduleContext} is decremented by the VM.
 *
 * @see ScheduleController
 */
public class ScheduleContext {

	/*
	 * WARNING: VM knows this class.
	 */
	/**
	 * Total number of quanta consumed by this context when last counter reset
	 * occurred. The current number of quanta consumed needs to be computed from
	 * this value and "counter" value.
	 */
	volatile private long consumedQuantum;// VM Knows this field (don't move it)

	/**
	 * Context unique id. Used for debug purpose.
	 */
	volatile private int uid;// VM Knows this field (don't move it)

	/**
	 * Maximum number of quantum that can be consumed by this context.
	 */
	volatile private int quota;// VM Knows this field (don't move it)

	/**
	 * Remaining quantum for this context.
	 */
	volatile private int counter;// VM Knows this field (don't move it)

	/**
	 * First element of the linked-list of threads attached to this context,
	 */
	volatile private int firstEngine;// VM Knows this field (don't move it)


	private ScheduleContext() {
		throw new RuntimeException();
	}

	/*default*/ ScheduleContext(int quota){
		this.initialize();
		this.setQuantumQuotaIntern(quota);
	}

	/**
	 * Set the time quota of the threads of this context. This quota is expressed in quanta of time.<p>
	 * A {@link ScheduleContext} with a quantum quota at 0 will never be scheduled.<p>
	 * A {@link ScheduleContext} with a quantum quota at -1 is always eligible to scheduling.<p>
	 * Calling this method induces a global reset of the quantum of all the contexts.
	 * @param quota the quota to set to this {@link ScheduleContext}.
	 *
	 * @throws IllegalArgumentException if the given quota is lower than -1.
	 */
	public void setQuantumQuota(int quota){
		//Set the quota and launch a quantum reset
		setQuantumQuotaIntern(quota);
		ScheduleController.resetQuantumOfAllContexts();
	}

	/**
	 * Just set the quota value
	 */
	private void setQuantumQuotaIntern(int quota){
		if(quota < -1){
			throw new IllegalArgumentException(Integer.toString(quota));
		}

		setQuantumQuotaNative(quota);
	}

	/**
	 * @return the quantum quota of this {@link ScheduleContext}.
	 */
	public int getQuantumQuota(){
		return quota;
	}

	/**
	 * Initializes this new created instance, and assigns its {@link #uid}.
	 */
	native public void initialize();

	/**
	 * @return the total amount of quantum that have been consumed by this
	 *         {@link ScheduleContext}.
	 */
	native public long getQuantumCounter();

	/**
	 * @param quota must be >= -1
	 */
	native public void setQuantumQuotaNative(int quota);

	/**
	 * Add the given thread in this {@link ScheduleContext}.
	 * @param thread the thread to add in the context.
	 * @return the previous context of the thread. Null if the thread was not associated to any context.
	 *
	 * @throws IllegalThreadStateException if the thread is not started.
	 */
	native public ScheduleContext addThread(Thread thread);

	/**
	 * Starts the given thread in this context.
	 * @param thread the thread to start
	 *
	 * @throws IllegalThreadStateException if the thread was already started.
	 *
	 * @see Thread#start()
	 */
	public void startThread(Thread thread){
		ScheduleController.getScheduleController().startThread(thread, this);
	}

}
